/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. See
 * http://www.eu-egee.org/partners/ for details on the copyright holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package org.glite.security.delegation.impl;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.security.GeneralSecurityException;
import java.security.KeyStoreException;
import java.security.PrivateKey;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import org.apache.log4j.Logger;
import org.bouncycastle.jce.PKCS10CertificationRequest;
import org.bouncycastle.openssl.PEMReader;
import org.bouncycastle.openssl.PEMWriter;
import org.glite.security.delegation.CertInfoTriple;
import org.glite.security.delegation.GrDPX509Util;
import org.glite.security.delegation.GrDProxyDlgeeOptions;
import org.glite.security.delegation.DelegationException;
import org.glite.security.delegation.NewProxyReq;
import org.glite.security.delegation.storage.GrDPStorage;
import org.glite.security.delegation.storage.GrDPStorageCacheElement;
import org.glite.security.delegation.storage.GrDPStorageElement;
import org.glite.security.delegation.storage.GrDPStorageException;
import org.glite.security.delegation.storage.GrDPStorageFactory;
import org.italiangrid.voms.VOMSValidators;
import org.italiangrid.voms.ac.VOMSACValidator;
import org.italiangrid.voms.store.VOMSTrustStore;
import org.italiangrid.voms.store.VOMSTrustStores;

import eu.emi.security.authn.x509.CrlCheckingMode;
import eu.emi.security.authn.x509.NamespaceCheckingMode;
import eu.emi.security.authn.x509.OCSPParametes;
import eu.emi.security.authn.x509.ProxySupport;
import eu.emi.security.authn.x509.RevocationParameters;
import eu.emi.security.authn.x509.StoreUpdateListener;
import eu.emi.security.authn.x509.ValidationError;
import eu.emi.security.authn.x509.ValidationErrorListener;
import eu.emi.security.authn.x509.RevocationParameters.RevocationCheckingOrder;
import eu.emi.security.authn.x509.impl.CertificateUtils;
import eu.emi.security.authn.x509.impl.CertificateUtils.Encoding;
import eu.emi.security.authn.x509.impl.KeyAndCertCredential;
import eu.emi.security.authn.x509.impl.OpensslCertChainValidator;
import eu.emi.security.authn.x509.impl.ValidatorParams;
import eu.emi.security.authn.x509.impl.X500NameUtils;
import eu.emi.security.authn.x509.proxy.ProxyCSR;
import eu.emi.security.authn.x509.proxy.ProxyCSRGenerator;
import eu.emi.security.authn.x509.proxy.ProxyCertificateOptions;
import eu.emi.security.authn.x509.proxy.ProxyChainInfo;
import eu.emi.security.authn.x509.proxy.ProxyType;

/**
 * Implementation of the server side logic of the Glite Delegation Interface.
 * 
 * @author Ricardo Rocha <Ricardo.Rocha@cern.ch>
 * @author Akos Frohner <Akos.Frohner@cern.ch>
 * @author Joni Hahkala <Joni.Hahkala@cern.ch>
 * 
 */
public class GliteDelegation {

    /** Local logger object. */
    public static Logger logger = Logger.getLogger(GliteDelegation.class);

    /**
     * The default key size to be used. Can be overwritten by setting the dlgeeKeySize property in the dlgee.properties
     * file if the value there is bigger.
     */
    private int DEFAULT_KEY_SIZE = 1024;

    /**
     * Set at instantiation time. Remains false if a bad configuration set was found.
     */
    private boolean m_bad_config = true;

    /** Local object interfacing the storage area. */
    private GrDPStorage m_storage = null;

    /** Key size being used. */
    private int m_keySize;

    /** whether the presence of voms attributes will be required in the incoming certificate chains. */
    private static boolean requireVomsAttrs = true;

    /** The voms validator instance used to validate the voms attribute certificates. */
    private static VOMSACValidator vomsValidator = null;
    
    /** Whether the new proxies should be limited or not. */
    private boolean limited = false;
    

    /**
     * Loads the DLGEE properties from the default config file and calls the appropriate constructor.
     * 
     * @throws IOException Failed to load the DLGEE config file
     * @see #GliteDelegation(GrDProxyDlgeeOptions)
     */
    public GliteDelegation() throws Exception {
        this(new GrDProxyDlgeeOptions(GrDPX509Util.getDlgeePropertyFile()));
    }

    /**
     * Class constructor.
     * 
     * Creates a new storage handler instance (implementation depending on configuration) to be used later.
     * 
     * Sets the value of the key size as defined in the configuration.
     * 
     * Creates the voms validator to be used for extracting the voms attributes from the given certificate chain.
     * 
     * @param dlgeeOpt the options object for configuring the delegation receiver.
     */
    public GliteDelegation(GrDProxyDlgeeOptions dlgeeOpt) throws Exception {

        // this.m_dlgeeOpt = dlgeeOpt;
        if (logger.isDebugEnabled()) {
            logger.debug("Using DLGEE properties: " + "DN: " + dlgeeOpt.getDlgeeDN() + ". Pass: <hidden>. proxyFile: "
                    + dlgeeOpt.getDlgeeProxyFile() + ". " + "delegationStorageFactory: "
                    + dlgeeOpt.getDlgeeStorageFactory());
        }
        // Get a GrDStorage instance
        try {
            GrDPStorageFactory stgFactory = GrDPX509Util.getGrDPStorageFactory(dlgeeOpt.getDlgeeStorageFactory());

            m_storage = stgFactory.createGrDPStorage(dlgeeOpt);
        } catch (Exception e) {
            logger.error("Failed to get a GrDPStorage instance. Delegation is not active.", e);
            m_bad_config = true;
            return;
        }

        requireVomsAttrs = dlgeeOpt.isRequireVomsAttributes();
        
        limited = dlgeeOpt.isLimited();

        // Set the size of the key, if not defined or smaller than the default,
        // use default.
        m_keySize = dlgeeOpt.getDlgeeKeySize();
        if (m_keySize == -1 || m_keySize < DEFAULT_KEY_SIZE) {
            m_keySize = DEFAULT_KEY_SIZE;
        }
        try {
            VOMSTrustStore vomsTrustStore = null;
            String vomsDirString = dlgeeOpt.getVomsDir();
            if (vomsDirString != null) {
                List<String> trustDirStrings = new ArrayList<String>();
                trustDirStrings.add(vomsDirString);
                vomsTrustStore = VOMSTrustStores.newTrustStore(trustDirStrings);
            } else {
                vomsTrustStore = VOMSTrustStores.newTrustStore();
            }

            StoreUpdateListener listener = new StoreUpdateListener() {
                public void loadingNotification(String location, String type, Severity level, Exception cause) {
                    if (level != Severity.NOTIFICATION) {
                        logger.error("Error when creating or using SSL socket. Type " + type + " level: " + level
                                + ((cause == null) ? "" : (" cause: " + cause.getClass() + ":" + cause.getMessage())));
                    } else {
                        // log successful (re)loading
                    }
                }
            };

            ArrayList<StoreUpdateListener> listenerList = new ArrayList<StoreUpdateListener>();
            listenerList.add(listener);

            RevocationParameters revParam = new RevocationParameters(CrlCheckingMode.REQUIRE, new OCSPParametes(),
                    false, RevocationCheckingOrder.CRL_OCSP);
            String crlCheckingMode = dlgeeOpt.getRevocationChecking();
            if (crlCheckingMode != null) {
                if (crlCheckingMode.equalsIgnoreCase("ifvalid")) {
                    revParam = new RevocationParameters(CrlCheckingMode.IF_VALID, new OCSPParametes(), false,
                            RevocationCheckingOrder.CRL_OCSP);
                }
                if (crlCheckingMode.equalsIgnoreCase("ignore")) {
                    revParam = new RevocationParameters(CrlCheckingMode.IGNORE, new OCSPParametes(), false,
                            RevocationCheckingOrder.CRL_OCSP);
                }
            }

            ValidatorParams validatorParams = new ValidatorParams(revParam, ProxySupport.ALLOW, listenerList);

            String trustStoreLocation = dlgeeOpt.getVomsCAs();
            if (trustStoreLocation == null) {
                trustStoreLocation = "/etc/grid-security/certificates";
            }

            String namespaceModeString = dlgeeOpt.getNamespace();
            NamespaceCheckingMode namespaceMode = NamespaceCheckingMode.EUGRIDPMA_AND_GLOBUS;
            if (namespaceModeString != null) {
                if (namespaceModeString.equalsIgnoreCase("no") || namespaceModeString.equalsIgnoreCase("false")
                        || namespaceModeString.equalsIgnoreCase("off")) {
                    namespaceMode = NamespaceCheckingMode.IGNORE;
                } else {
                    if (namespaceModeString.equalsIgnoreCase("require")) {
                        namespaceMode = NamespaceCheckingMode.EUGRIDPMA_AND_GLOBUS_REQUIRE;
                    }
                }

            }

            String intervalString = dlgeeOpt.getUpdateInterval();
            long intervalMS = 3600000; // update every hour
            if (intervalString != null) {
                intervalMS = Long.parseLong(intervalString);
            }

            OpensslCertChainValidator validator = new OpensslCertChainValidator(trustStoreLocation, namespaceMode,
                    intervalMS, validatorParams);

            ValidationErrorListener validationListener = new ValidationErrorListener() {
                @Override
                public boolean onValidationError(ValidationError error) {
                    logger.info("Error when validating incoming certificate: " + error.getMessage() + " position: "
                            + error.getPosition() + " " + error.getParameters());
                    X509Certificate chain[] = error.getChain();
                    for (X509Certificate cert : chain) {
                        logger.info(cert.toString());
                    }
                    return false;
                }

            };

            validator.addValidationListener(validationListener);

            vomsValidator = VOMSValidators.newValidator(vomsTrustStore, validator);
        } catch (Exception e) {
            if (requireVomsAttrs) {
                logger.error("Error while creating VOMS validator: " + e.getMessage(), e);
                throw e;
            } else {
                logger.warn("Error while creating VOMS validator: " + e.getMessage(), e);
            }
        }

        // set the config ok flag
        m_bad_config = false;
    }

    /**
     * Generates a new proxy request based on the certificates and the possibly given delegation id. Also checks if a
     * delegation with given (or generated if not given) id for the user already exists. If a delegation with given id
     * and user already exists, it throws a delegation exception. To renew existing delegation, use renewProxyReq.
     * 
     * @param inDelegationID The delegation id to use (generated if not present, based on dn and possible VOMS
     *            attributes).
     * @param certs The certificates the user used to authenticate himself.
     * @return The generated Proxy request in PEM encoding.
     * @throws DelegationException Thrown in case of failures.
     */
    public String getProxyReq(String inDelegationID, X509Certificate certs[]) throws DelegationException {
        logger.debug("Processing getProxyReq.");

        if (certs == null || certs.length == 0) {
            logger.error("Did not get any certificates.");
            throw new DelegationException("Did not get any certificates.");
        }

        logger.debug("Got " + certs.length + " certs.");
        String delegationID = inDelegationID;

        GrDPStorageElement elem = null;

        // Check if a bad configuration was detected on launch (and fail if
        // true)
        if (m_bad_config) {
            logger.error("Service is misconfigured. Stopping execution.");
            throw new DelegationException("Service is misconfigured.");
        }
        CertInfoTriple info = null;
        try {
            info = new CertInfoTriple(certs, vomsValidator, requireVomsAttrs);
        } catch (Exception e) {
            logger.error("Getting info from the certificate chain failed: " + e.getClass() + " " + e.getMessage(), e);
            throw new DelegationException("Getting info from the certificate chain failed: " + e.getClass() + " "
                    + e.getMessage(), e);
        }

        logger.debug("Got get proxy req request from client '" + info.dn + "', got VOMS attributes : "
                + info.vomsAttributes);
        if (info.vomsAttributes != null) {
            for (String attrib : info.vomsAttributes) {
                logger.debug("Voms attrib : " + attrib);
            }
        }

        // Generate a delegation id from the client DN and VOMS attributes
        if (delegationID == null || delegationID.length() == 0) {
            delegationID = GrDPX509Util.genDlgID(info.dn, info.vomsAttributes);
        }

        logger.debug("Delegation id is: " + delegationID);

        // Search for an existing entry in storage for this delegation ID (null
        // if non existing)
        try {
            elem = m_storage.findGrDPStorageElement(delegationID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Failure on storage interaction.", e);
            throw new DelegationException("Internal failure.");
        }

        // Throw error in case there was already a credential with the given id
        if (elem != null) {
            String vomsAttrsStr = GrDPX509Util.toStringVOMSAttrs(info.vomsAttributes);
            logger.debug("Delegation ID '" + delegationID + "' already exists" + " for client (DN='" + info.dn
                    + "; VOMS ATTRS='" + vomsAttrsStr + "'). Call renewProxyReq.");
            throw new DelegationException("Delegation ID '" + delegationID + "' already exists" + " for client (DN='"
                    + info.dn + "; VOMS ATTRS='" + vomsAttrsStr + "'). Call renewProxyReq.");
        }

        // Create and store the new certificate request
        return createAndStoreCertificateRequest(certs, delegationID, info.dn, info.vomsAttributes);
    }

    /**
     * Generates a new proxy request based on the certificates and the possibly given delegation id. Also checks if a
     * delegation with given (or generated if not given) id for the user already exists. If a delegation with given id
     * and user already exists, it throws a delegation exception. To renew existing delegation, use renewProxyReq.
     * 
     * This method also returns the given or generated delegation ID.
     * 
     * @param inDelegationID the delegation id to use, will be generated (based on dn and VOMS attributes if present) if
     *            not given.
     * @param certs The certificates the user used to authenticate himself.
     * @return The newProxyReq object.
     * @throws DelegationException thrown in case of failure.
     */
    public NewProxyReq getNewProxyReq(String inDelegationID, X509Certificate certs[]) throws DelegationException {
        logger.debug("Processing getNewProxyReq.");

        String delegationID = inDelegationID;

        GrDPStorageElement elem = null;

        // Check if a bad configuration was detected on launch (and fail if
        // true)
        if (m_bad_config) {
            logger.error("Service is misconfigured. Stopping execution.");
            throw new DelegationException("Service is misconfigured.");
        }

        CertInfoTriple info = new CertInfoTriple(certs, vomsValidator, requireVomsAttrs);

        logger.debug("Got get new proxy req request from client '" + info.dn + "'");

        // Generate a delegation id from the client DN and VOMS attributes
        if (delegationID == null || delegationID.length() == 0) {
            delegationID = GrDPX509Util.genDlgID(info.dn, info.vomsAttributes);
        }

        // Search for an existing entry in storage for this delegation ID (null
        // if non existing)
        try {
            elem = m_storage.findGrDPStorageElement(delegationID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Failure on storage interaction.", e);
            throw new DelegationException("Internal failure.");
        }

        // Throw error in case there was already a credential with the given id
        if (elem != null) {
            String vomsAttrsStr = GrDPX509Util.toStringVOMSAttrs(info.vomsAttributes);
            String errorMsg = "Delegation ID '" + delegationID + "' already exists" + " for client (DN='" + info.dn
                    + "; VOMS ATTRS='" + vomsAttrsStr + "'). Call renewProxyReq.";

            logger.debug(errorMsg);
            throw new DelegationException(errorMsg);
        }

        // Create and store the new certificate request
        String certRequest = createAndStoreCertificateRequest(certs, delegationID, info.dn, info.vomsAttributes);

        // Create and return the proxy request object
        NewProxyReq newProxyReq = new NewProxyReq();
        newProxyReq.setDelegationID(delegationID);
        newProxyReq.setProxyRequest(certRequest);

        return newProxyReq;
    }

    /**
     * Generates a new delegation request for the existing delegation with the given (or generated) delegation.
     * 
     * @param inDelegationID The delegation id to use, will be generated (based on dn and VOMS attributes if present) if
     *            not given.
     * @param certs The certificates the user used to authenticate himself.
     * @return The delegation request in PEM format.
     * @throws DelegationException Thrown in case of failure.
     */
    public String renewProxyReq(String inDelegationID, X509Certificate certs[]) throws DelegationException {
        logger.debug("Processing renewProxyReq.");

        String delegationID = inDelegationID;

        GrDPStorageElement elem = null;

        // Check if a bad configuration was detected on launch (and fail if
        // true)
        if (m_bad_config) {
            logger.error("Service is misconfigured. Stopping execution.");
            throw new DelegationException("Service is misconfigured.");
        }

        CertInfoTriple info = new CertInfoTriple(certs, vomsValidator, requireVomsAttrs);

        logger.debug("Got renew proxy request from client '" + info.dn + "'");

        // Generate a delegation id from the client DN and VOMS attributes
        if (delegationID == null || delegationID.length() == 0) {
            delegationID = GrDPX509Util.genDlgID(info.dn, info.vomsAttributes);
        }

        // Search for an existing entry in storage for this delegation ID (null
        // if non existing)
        try {
            elem = m_storage.findGrDPStorageElement(delegationID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Failure on storage interaction.", e);
            throw new DelegationException("Internal failure.");
        }

        // Check that the DLG ID had a corresponding delegated credential
        if (elem == null) {
            logger.debug("Failed to renew credential as there was no delegation with ID '" + delegationID
                    + "' for client '" + info.dn + "'");
        }

        // Create and store the new certificate request
        return createAndStoreCertificateRequest(certs, delegationID, info.dn, info.vomsAttributes);
    }

    /**
     * Puts the new proxy that was signed by the client to the delegation storage. Also fetches the private key from the
     * cache so that the proxy credentials are complete.
     * 
     * @param inDelegationID The delegation ID used for the delegation.
     * @param proxy The new proxy received from the client.
     * @param certs The certificates the user used to authenticate himself.
     * @throws DelegationException
     */
    public void putProxy(String inDelegationID, String proxy, X509Certificate certs[]) throws Exception {// throws
                                                                                                         // DelegationException
                                                                                                         // {
        logger.info("Processing putProxy.");

        String delegationID = inDelegationID;

        // Check if a bad configuration was detected on launch (and fail if
        // true)
        if (m_bad_config) {
            logger.error("Service is misconfigured. Stopping execution.");
            throw new DelegationException("Service is misconfigured.");
        }

        // Check for a null proxy
        if (proxy == null) {
            logger.error("Failed to putProxy as proxy was null.");
            throw new DelegationException("No proxy was given.");
        }

        CertInfoTriple info = new CertInfoTriple(certs, vomsValidator, requireVomsAttrs);

        logger.debug("Got put proxy request from client '" + info.dn + "'");

        // Load given proxy
        X509Certificate[] proxyCertChain;
        try {
            proxyCertChain = CertificateUtils.loadCertificateChain(new BufferedInputStream(new ByteArrayInputStream(
                    proxy.getBytes())), Encoding.PEM);
        } catch (IOException e2) {
            logger.error("Failed to load proxy certificate chain: " + e2.getMessage());
            throw new DelegationException("Failed to load proxy certificate chain: " + e2.getMessage());
        }
        if (proxyCertChain == null || proxyCertChain.length == 0) {
            logger.error("Failed to load proxy certificate chain - chain was null or size 0.");
            throw new DelegationException("Failed to load proxy certificate chain.");
        }
        logger.debug("Given proxy certificate loaded successfully.");

        // check if the chain is within it's validity period.
        for (int i = 0; i < proxyCertChain.length; i++) {
            // Check if the proxy is currently valid
            try {
                proxyCertChain[i].checkValidity();
            } catch (CertificateExpiredException e) {
                throw new DelegationException("Failed proxy validation - it expired on: "
                        + proxyCertChain[0].getNotAfter());
            } catch (CertificateNotYetValidException e) {
                throw new DelegationException("Failed proxy validation - it will be valid from: "
                        + proxyCertChain[0].getNotBefore());
            }
        }

        // Get the given proxy information
        String proxySubjectDN = X500NameUtils.getReadableForm(proxyCertChain[0].getSubjectX500Principal());
        String proxyIssuerDN = X500NameUtils.getReadableForm(proxyCertChain[0].getIssuerX500Principal());
        if (logger.isDebugEnabled()) {
            logger.debug("Proxy Subject DN: " + proxySubjectDN);
            logger.debug("Proxy Issuer DN: " + proxyIssuerDN);
            logger.debug("Proxy Public key:" + proxyCertChain[0].getPublicKey());
            logger.debug("chain length is: " + proxyCertChain.length);
            logger.debug("last cert is:" + proxyCertChain[proxyCertChain.length - 1]);

            for (int n = 0; n < proxyCertChain.length; n++) {
                logger.debug("cert [" + n + "] is from "
                        + X500NameUtils.getReadableForm(proxyCertChain[n].getSubjectX500Principal()));
            }
        }

        if (proxySubjectDN == null || proxyIssuerDN == null) {
            logger.error("Failed to get DN (subject or issuer) out of proxy. It came null");
            throw new DelegationException("Failed to get DN (subject or issuer) out of proxy.");
        }

        logger.debug("Client DN: " + info.dn);

        // Get a delegation ID for the given proxy (or take the specified one ifgiven)
        // TODO: Should the dlg id here be generated from the client or theproxy info?
        // Also, should the client and proxy VOMS attributes be checked for amatch?
        if (delegationID == null || delegationID.length() == 0) {
            delegationID = GrDPX509Util.genDlgID(info.dn, info.vomsAttributes);
        }
        logger.debug("Delegation ID is '" + delegationID + "'");

        // Check that the client is the issuer of the given proxy
        // TODO: more strict check
        if (!proxyIssuerDN.endsWith(info.dn)) {
            String message = "Client '" + info.dn + "' is not issuer of proxy '" + proxyIssuerDN + "'.";
            logger.error(message);
            throw new DelegationException(message);
        }

        String cacheID = delegationID;
        cacheID = delegationID + '+' + GrDPX509Util.generateSessionID(proxyCertChain[0].getPublicKey());
        logger.debug("public key is: " + proxyCertChain[0].getPublicKey());
        logger.debug("Cache ID (delegation ID + session ID): " + cacheID);

        // Get the cache entry for this delegation ID
        GrDPStorageCacheElement cacheElem = null;
        try {
            cacheElem = m_storage.findGrDPStorageCacheElement(cacheID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Failed to get certificate request information from storage.", e);
            throw new DelegationException("Internal failure.");
        }

        // Check if the delegation request existed
        if (cacheElem == null) {
            logger.info("Could not find cache ID '" + cacheID + "' for DN '" + info.dn + "' in cache.");
            throw new DelegationException("Could not find a proper delegation request");
        }
        logger.debug("Got from cache element for cache ID '" + cacheID + "' and DN '" + info.dn + "'");

        // the public key of the cached certificate request has to
        // match the public key of the proxy certificate, otherwise
        // this is an answer to a different request
        PEMReader pemReader = new PEMReader(new StringReader(cacheElem.getCertificateRequest()));
        PKCS10CertificationRequest req;
        try {
            req = (PKCS10CertificationRequest) pemReader.readObject();
            pemReader.close();
        } catch (IOException e1) {
            logger.error("Could not load the original certificate request from cache.");
            throw new DelegationException("Could not load the original certificate request from cache: "
                    + e1.getMessage());
        }
        if (req == null) {
            logger.error("Could not load the original certificate request from cache.");
            throw new DelegationException("Could not load the original certificate request from cache.");
        }
        try {
            if (!req.getPublicKey().equals(proxyCertChain[0].getPublicKey())) {
                logger.error("The proxy and the original request's public key do not match.");
                logger.error("Proxy public key: " + proxyCertChain[0].getPublicKey());
                logger.error("Request public key: " + req.getPublicKey());
                throw new DelegationException("The proxy and the original request's public key do not match.");
            }
        } catch (GeneralSecurityException ge) {
            logger.error("Error while decoding the certificate request: " + ge);
            throw new DelegationException("Error while decoding the certificate request.");
        }

        // Add the private key to the proxy certificate chain and check it was ok
        ByteArrayInputStream keyStream = new ByteArrayInputStream(cacheElem.getPrivateKey().getBytes());
        PrivateKey privateKey;
        try {
            privateKey = CertificateUtils.loadPrivateKey(keyStream, Encoding.PEM, null);
        } catch (IOException e) {
            throw new DelegationException("Failed to read private key from storage, error: " + e.getClass() + ": "
                    + e.getMessage());
        }
        KeyAndCertCredential credential;
        try {
            credential = new KeyAndCertCredential(privateKey, proxyCertChain);
        } catch (KeyStoreException e) {
            throw new DelegationException("Failed to handle credentials, error: " + e.getClass() + ": "
                    + e.getMessage());
        }

        ByteArrayOutputStream proxyStream = new ByteArrayOutputStream();
        try {
            CertificateUtils.savePEMKeystore(proxyStream, credential.getKeyStore(), credential.getKeyAlias(), null,
                    credential.getKeyPassword(), null, true);
        } catch (Exception e) {
            logger.error("Error while converting the proxy to string for storage: " + e.getClass() + ": "
                    + e.getMessage());
            e.printStackTrace();
            throw new DelegationException("Error while converting the proxy to string for storage: " + e.getClass()
                    + ": " + e.getMessage());
        }

        String completeProxy = proxyStream.toString();

        if (completeProxy == null) {
            logger.error("Failed to add private key to the proxy certificate chain.");
            throw new DelegationException("Could not properly process given proxy.");
        }

        // Save the proxy in proxy storage (copying the rest from the info taken from the cache)
        try {
            GrDPStorageElement elem = m_storage.findGrDPStorageElement(delegationID, info.dn);
            if (elem != null) {
                elem.setCertificate(completeProxy);
                elem.setTerminationTime(proxyCertChain[0].getNotAfter());
                m_storage.updateGrDPStorageElement(elem);
            } else {
                elem = new GrDPStorageElement();
                elem.setDelegationID(delegationID);
                elem.setDN(info.dn);
                elem.setVomsAttributes(info.vomsAttributes);
                elem.setCertificate(completeProxy);
                elem.setTerminationTime(proxyCertChain[0].getNotAfter());
                m_storage.insertGrDPStorageElement(elem);
            }
        } catch (GrDPStorageException e) {
            logger.error("Failed to put certificate request in storage.", e);
            throw new DelegationException("Internal failure: " + e.getMessage());
        }
        logger.debug("Delegation finished successfully.");

        // Remove the credential from storage cache
        try {
            m_storage.deleteGrDPStorageCacheElement(cacheID, info.dn);
        } catch (GrDPStorageException e) {
            logger.warn("Failed to remove credential from storage cache.");
        }

    }

    public void destroy(String inDelegationID, X509Certificate certs[]) throws DelegationException {
        logger.debug("Processing destroy.");

        String delegationID = inDelegationID;

        GrDPStorageElement elem = null;

        // Check if a bad configuration was detected on launch (and fail if
        // true)
        if (m_bad_config) {
            logger.error("Service is misconfigured. Stopping execution.");
            throw new DelegationException("Service is misconfigured.");
        }

        CertInfoTriple info = new CertInfoTriple(certs, vomsValidator, requireVomsAttrs);

        // Generate a delegation id from the client DN and VOMS attributes
        if (delegationID == null || delegationID.length() == 0) {
            delegationID = GrDPX509Util.genDlgID(info.dn, info.vomsAttributes);
        }

        logger.debug("Got destroy request for delegation id '" + delegationID + "' from client '" + info.dn + "'");

        // Search for an existing entry in storage for this delegation ID (null
        // if non existing)
        try {
            elem = m_storage.findGrDPStorageElement(delegationID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Failure on storage interaction. Exception: ", e);
            throw new DelegationException("Internal failure.");
        }

        // Throw exception if non-existing
        if (elem == null) {
            logger.debug("Failed to find delegation ID '" + delegationID + "' for client '" + info.dn + "' in storage.");
            throw new DelegationException("Failed to find delegation ID '" + delegationID + "' in storage.");
        }

        // Remove the credential from storage
        try {
            m_storage.deleteGrDPStorageElement(delegationID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Inconsistency needs manual intervention. Delegation ID '" + delegationID + " of client '"
                    + info.dn + "' was found, " + "but could not be removed from storage.");
            throw new DelegationException("Failed to destroy delegated credential.");
        }

        logger.debug("Delegated credential destroyed.");
    }

    public Calendar getTerminationTime(String inDelegationID, X509Certificate certs[]) throws DelegationException {
        logger.debug("Processing getTerminationTime.");

        String delegationID = inDelegationID;

        GrDPStorageElement elem = null;

        // Check if a bad configuration was detected on launch (and fail if
        // true)
        if (m_bad_config) {
            logger.error("Service is misconfigured. Stopping execution.");
            throw new DelegationException("Service is misconfigured.");
        }

        CertInfoTriple info = new CertInfoTriple(certs, vomsValidator, requireVomsAttrs);

        // Generate a delegation id from the client DN and VOMS attributes
        if (delegationID == null || delegationID.length() == 0) {
            delegationID = GrDPX509Util.genDlgID(info.dn, info.vomsAttributes);
        }

        logger.debug("Got getTerminationTime request for delegation id '" + delegationID + "' from client '" + info.dn
                + "'");

        // Search for an existing entry in storage for this delegation ID (null if non existing)
        try {
            elem = m_storage.findGrDPStorageElement(delegationID, info.dn);
        } catch (GrDPStorageException e) {
            logger.error("Failure on storage interaction. Exception: ", e);
            throw new DelegationException("Internal failure.");
        }

        // Throw exception if non-existing
        if (elem == null) {
            logger.debug("Failed to find delegation ID '" + delegationID + "' for client '" + info.dn + "' in storage.");
            throw new DelegationException("Failed to find delegation ID '" + delegationID + "' in storage.");
        }

        // Build a calendar object with the proper time
        Calendar cal = Calendar.getInstance();
        cal.setTime(elem.getTerminationTime());

        return cal;
    }

    /**
     * Creates a new certificate request and stores it in the storage cache area.
     * 
     * @param certs The certificates the user used to authenticate himself.
     * @param dlgID The delegation ID of the new delegation
     * @param clientDN The DN of the owner of the delegated credential
     * @param vomsAttributes The list of VOMS attributes in the delegated credential
     * @return The certificate request for the new delegated credential
     * @throws DelegationException Failed to create or store the new credential request
     */
    private String createAndStoreCertificateRequest(X509Certificate certs[], String dlgID, String clientDN,
            String[] vomsAttributes) throws DelegationException {

        // Generate the certificate request
        String certRequest = null;
        PrivateKey privKey = null;
        ProxyCSR proxyCsr = null;
        String privateKeyString = null;
        try {
            ProxyCertificateOptions options = new ProxyCertificateOptions(certs);
            ProxyChainInfo info = new ProxyChainInfo(certs);
            ProxyType type = info.getProxyType().toProxyType();
            options.setType(type);
            options.setKeyLength(m_keySize);
            if(limited || info.isLimited()){
//                if(type != ProxyType.LEGACY){
//                    options.setPolicy(new ProxyPolicy(ProxyPolicy.LIMITED_PROXY_OID));
//                }
                options.setLimited(limited);
            }
            proxyCsr = ProxyCSRGenerator.generate(options);
            PKCS10CertificationRequest req = proxyCsr.getCSR();
            privKey = proxyCsr.getPrivateKey();
            // System.out.println(req.getCertificationRequestInfo().getSubject());
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            try {
                CertificateUtils.savePrivateKey(stream, privKey, Encoding.PEM, null, null);
            } catch (IllegalArgumentException e) {
                logger.error("Failed to generate certificate request.", e);
                throw new DelegationException("Internal failure, failed to generate certificate request.", e);
            } catch (IOException e) {
                logger.error("Failed to generate certificate request.", e);
                throw new DelegationException("Internal failure, failed to generate certificate request.", e);
            }
            privateKeyString = stream.toString();

            StringWriter stringWriter = new StringWriter();
            PEMWriter pemWriter = new PEMWriter(stringWriter);
            try {
                pemWriter.writeObject(req);
                pemWriter.flush();
                pemWriter.close();
            } catch (IOException e) {
                logger.error("Failed to put certificate request in storage.", e);
                throw new DelegationException("Internal failure.", e);
            }

            certRequest = stringWriter.toString();
        } catch (Exception e) {
            logger.error("Error while generating the certificate request." + e);
            throw new DelegationException("Failed to generate a certificate request. " + e.getMessage());
        }
        logger.debug("Certificate request generation was successfull.");

        String cacheID = null;
        try {
            cacheID = dlgID + '+' + GrDPX509Util.generateSessionID(proxyCsr.getCSR().getPublicKey());
            logger.debug("public key is: " + proxyCsr.getCSR().getPublicKey());
            logger.debug("Cache ID (delegation ID + session ID): " + cacheID);
        } catch (Exception e) {
            logger.error("Failed to generate certificate request.", e);
            throw new DelegationException("Internal failure, failed to generate certificate request.", e);
        }

        try {
            GrDPStorageCacheElement cacheElem = new GrDPStorageCacheElement();
            cacheElem.setDelegationID(cacheID);
            cacheElem.setDN(clientDN);
            cacheElem.setVomsAttributes(vomsAttributes);
            cacheElem.setCertificateRequest(certRequest);
            cacheElem.setPrivateKey(privateKeyString);
            m_storage.insertGrDPStorageCacheElement(cacheElem);
        } catch (GrDPStorageException e) {
            logger.error("Failed to put certificate request in storage.", e);
            throw new DelegationException("Internal failure.", e);
        }
        logger.debug("New certificate request successfully stored in cache.");

        return certRequest;
    }

}
